@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright 2018 Data61, CSIRO
+# Copyright 2018-2020 Data61, CSIRO
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -26,5 +26,15 @@ from .graphsage import *
 from .hinsage import *
 from .graph_attention import *
 from .link_inference import *
+from .ppnp import *
+from .appnp import *
 from .gcn import *
+from .cluster_gcn import *
+from .attri2vec import *
 from .misc import SqueezedSparseConversion
+from .preprocessing_layer import GraphPreProcessingLayer
+from .rgcn import *
+from .watch_your_step import *
+from .knowledge_graph import *
+from .graph_classification import *
+from .deep_graph_infomax import *
