@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright 2018-2019 Data61, CSIRO
+# Copyright 2018-2020 Data61, CSIRO
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,12 +14,14 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from keras import backend as K
-from keras import activations, initializers, constraints, regularizers
-from keras.layers import Input, Layer, Lambda, Dropout, Reshape
+import warnings
+from tensorflow.keras import backend as K
+from tensorflow.keras import activations, initializers, constraints, regularizers
+from tensorflow.keras.layers import Input, Layer, Lambda, Dropout, Reshape
 
-from ..mapper import FullBatchNodeGenerator
-from .misc import SqueezedSparseConversion
+from ..mapper import FullBatchGenerator
+from .misc import SqueezedSparseConversion, deprecated_model_function
+from .preprocessing_layer import GraphPreProcessingLayer
 
 
 class GraphConvolution(Layer):
@@ -51,18 +53,16 @@ class GraphConvolution(Layer):
 
     Args:
         units (int): dimensionality of output feature vectors
-        activation (str): nonlinear activation applied to layer's output to obtain output features
+        activation (str or func): nonlinear activation applied to layer's output to obtain output features
         use_bias (bool): toggles an optional bias
         final_layer (bool): If False the layer returns output for all nodes,
                             if True it returns the subset specified by the indices passed to it.
-        kernel_initializer (str): name of layer bias f the initializer for kernel parameters (weights)
-        bias_initializer (str): name of the initializer for bias
-        attn_kernel_initializer (str): name of the initializer for attention kernel
-        kernel_regularizer (str): name of regularizer to be applied to layer kernel. Must be a Keras regularizer.
-        bias_regularizer (str): name of regularizer to be applied to layer bias. Must be a Keras regularizer.
-        activity_regularizer (str): not used in the current implementation
-        kernel_constraint (str): constraint applied to layer's kernel
-        bias_constraint (str): constraint applied to layer's bias
+        kernel_initializer (str or func, optional): The initialiser to use for the weights.
+        kernel_regularizer (str or func, optional): The regulariser to use for the weights.
+        kernel_constraint (str or func, optional): The constraint to use for the weights.
+        bias_initializer (str or func, optional): The initialiser to use for the bias.
+        bias_regularizer (str or func, optional): The regulariser to use for the bias.
+        bias_constraint (str or func, optional): The constraint to use for the bias.
     """
 
     def __init__(
@@ -71,32 +71,32 @@ class GraphConvolution(Layer):
         activation=None,
         use_bias=True,
         final_layer=False,
+        input_dim=None,
         kernel_initializer="glorot_uniform",
-        bias_initializer="zeros",
         kernel_regularizer=None,
-        bias_regularizer=None,
-        activity_regularizer=None,
         kernel_constraint=None,
+        bias_initializer="zeros",
+        bias_regularizer=None,
         bias_constraint=None,
-        **kwargs,
+        **kwargs
     ):
-        if "input_shape" not in kwargs and "input_dim" in kwargs:
-            kwargs["input_shape"] = (kwargs.get("input_dim"),)
-
-        super().__init__(**kwargs)
+        if "input_shape" not in kwargs and input_dim is not None:
+            kwargs["input_shape"] = (input_dim,)
 
         self.units = units
         self.activation = activations.get(activation)
         self.use_bias = use_bias
-        self.kernel_initializer = initializers.get(kernel_initializer)
-        self.bias_initializer = initializers.get(bias_initializer)
-        self.kernel_regularizer = regularizers.get(kernel_regularizer)
-        self.bias_regularizer = regularizers.get(bias_regularizer)
-        self.activity_regularizer = regularizers.get(activity_regularizer)
-        self.kernel_constraint = constraints.get(kernel_constraint)
-        self.bias_constraint = constraints.get(bias_constraint)
         self.final_layer = final_layer
 
+        self.kernel_initializer = initializers.get(kernel_initializer)
+        self.kernel_regularizer = regularizers.get(kernel_regularizer)
+        self.kernel_constraint = constraints.get(kernel_constraint)
+        self.bias_initializer = initializers.get(bias_initializer)
+        self.bias_regularizer = regularizers.get(bias_regularizer)
+        self.bias_constraint = constraints.get(bias_constraint)
+
+        super().__init__(**kwargs)
+
     def get_config(self):
         """
         Gets class configuration for Keras serialization.
@@ -112,11 +112,10 @@ class GraphConvolution(Layer):
             "final_layer": self.final_layer,
             "activation": activations.serialize(self.activation),
             "kernel_initializer": initializers.serialize(self.kernel_initializer),
-            "bias_initializer": initializers.serialize(self.bias_initializer),
             "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
-            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
-            "activity_regularizer": regularizers.serialize(self.activity_regularizer),
             "kernel_constraint": constraints.serialize(self.kernel_constraint),
+            "bias_initializer": initializers.serialize(self.bias_initializer),
+            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
             "bias_constraint": constraints.serialize(self.bias_constraint),
         }
 
@@ -129,7 +128,7 @@ class GraphConvolution(Layer):
         Assumes the following inputs:
 
         Args:
-            input_shape (tuple of ints)
+            input_shapes (tuple of ints)
                 Shape tuples can include None for free dimensions, instead of an integer.
 
         Returns:
@@ -143,18 +142,18 @@ class GraphConvolution(Layer):
         else:
             out_dim = feature_shape[1]
 
-        return (batch_dim, out_dim, self.units)
+        return batch_dim, out_dim, self.units
 
     def build(self, input_shapes):
         """
         Builds the layer
 
         Args:
-            input_shape (list of int): shapes of the layer's inputs (node features and adjacency matrix)
+            input_shapes (list of int): shapes of the layer's inputs (node features and adjacency matrix)
 
         """
         feat_shape = input_shapes[0]
-        input_dim = feat_shape[-1]
+        input_dim = int(feat_shape[-1])
 
         self.kernel = self.add_weight(
             shape=(input_dim, self.units),
@@ -163,6 +162,7 @@ class GraphConvolution(Layer):
             regularizer=self.kernel_regularizer,
             constraint=self.kernel_constraint,
         )
+
         if self.use_bias:
             self.bias = self.add_weight(
                 shape=(self.units,),
@@ -207,7 +207,7 @@ class GraphConvolution(Layer):
         output = K.dot(h_graph, self.kernel)
 
         # Add optional bias & apply activation
-        if self.bias:
+        if self.bias is not None:
             output += self.bias
         output = self.activation(output)
 
@@ -233,10 +233,11 @@ class GCN:
     activation functions for each hidden layers, and a generator object.
 
     To use this class as a Keras model, the features and pre-processed adjacency matrix
-    should be supplied using the :class:`FullBatchNodeGenerator` class. To have the appropriate
-    pre-processing the generator object should be instantiated as follows::
+    should be supplied using either the :class:`FullBatchNodeGenerator` class for node inference
+    or the :class:`FullBatchLinkGenerator` class for link inference.
 
-        generator = FullBatchNodeGenerator(G, method="gcn")
+    To have the appropriate pre-processing the generator object should be instanciated
+    with the `method='gcn'` argument.
 
     Note that currently the GCN class is compatible with both sparse and dense adjacency
     matrices and the :class:`FullBatchNodeGenerator` will default to sparse.
@@ -244,6 +245,19 @@ class GCN:
     For more details, please see the GCN demo notebook:
     demos/node-classification/gat/gcn-cora-node-classification-example.ipynb
 
+    Example:
+        Creating a GCN node classification model from an existing :class:`StellarGraph`
+        object ``G``::
+
+            generator = FullBatchNodeGenerator(G, method="gcn")
+            gcn = GCN(
+                    layer_sizes=[32, 4],
+                    activations=["elu","softmax"],
+                    generator=generator,
+                    dropout=0.5
+                )
+            x_inp, predictions = gcn.in_out_tensors()
+
     Notes:
       - The inputs are tensors with a batch dimension of 1. These are provided by the \
         :class:`FullBatchNodeGenerator` object.
@@ -257,68 +271,82 @@ class GCN:
         However, the intermediate layers before the final layer order the nodes
         in the same way as the adjacency matrix.
 
-    Examples:
-        Creating a GCN node classification model from an existing :class:`StellarGraph`
-        object ``G``::
-
-            generator = FullBatchNodeGenerator(G, method="gcn")
-            gcn = GCN(
-                    layer_sizes=[32, 4],
-                    activations=["elu","softmax"],
-                    generator=generator,
-                    dropout=0.5
-                )
-            x_inp, predictions = gcn.node_model()
-
     Args:
-        layer_sizes (list of int): list of output sizes of GCN layers in the stack
-        activations (list of str): list of activations applied to each layer's output
-        generator (FullBatchNodeGenerator): an instance of FullBatchNodeGenerator class constructed on the graph of interest
-        bias (bool): toggles an optional bias in GCN layers
-        dropout (float): dropout rate applied to input features of each GCN layer
-        kernel_regularizer (str): normalization applied to the kernels of GCN layers
+        layer_sizes (list of int): Output sizes of GCN layers in the stack.
+        generator (FullBatchNodeGenerator): The generator instance.
+        bias (bool): If True, a bias vector is learnt for each layer in the GCN model.
+        dropout (float): Dropout rate applied to input features of each GCN layer.
+        activations (list of str or func): Activations applied to each layer's output;
+            defaults to ['relu', ..., 'relu'].
+        kernel_initializer (str or func, optional): The initialiser to use for the weights of each layer.
+        kernel_regularizer (str or func, optional): The regulariser to use for the weights of each layer.
+        kernel_constraint (str or func, optional): The constraint to use for the weights of each layer.
+        bias_initializer (str or func, optional): The initialiser to use for the bias of each layer.
+        bias_regularizer (str or func, optional): The regulariser to use for the bias of each layer.
+        bias_constraint (str or func, optional): The constraint to use for the bias of each layer.
     """
 
     def __init__(
         self,
         layer_sizes,
-        activations,
         generator,
         bias=True,
         dropout=0.0,
+        activations=None,
+        kernel_initializer="glorot_uniform",
         kernel_regularizer=None,
+        kernel_constraint=None,
+        bias_initializer="zeros",
+        bias_regularizer=None,
+        bias_constraint=None,
     ):
-        if not isinstance(generator, FullBatchNodeGenerator):
-            raise TypeError("Generator should be a instance of FullBatchNodeGenerator")
-
-        assert len(layer_sizes) == len(activations)
+        if not isinstance(generator, FullBatchGenerator):
+            raise TypeError(
+                "Generator should be a instance of FullBatchNodeGenerator or FullBatchLinkGenerator"
+            )
 
+        n_layers = len(layer_sizes)
         self.layer_sizes = layer_sizes
         self.activations = activations
         self.bias = bias
         self.dropout = dropout
-        self.kernel_regularizer = kernel_regularizer
-        self.generator = generator
-        self.support = 1
+
+        # Copy required information from generator
         self.method = generator.method
+        self.multiplicity = generator.multiplicity
+        self.n_nodes = generator.features.shape[0]
+        self.n_features = generator.features.shape[1]
 
         # Check if the generator is producing a sparse matrix
         self.use_sparse = generator.use_sparse
+        if self.method == "none":
+            self.graph_norm_layer = GraphPreProcessingLayer(num_of_nodes=self.n_nodes)
+
+        # Activation function for each layer
+        if activations is None:
+            activations = ["relu"] * n_layers
+        elif len(activations) != n_layers:
+            raise ValueError(
+                "Invalid number of activations; require one function per layer"
+            )
+        self.activations = activations
 
         # Initialize a stack of GCN layers
-        n_layers = len(self.layer_sizes)
         self._layers = []
         for ii in range(n_layers):
-            l = self.layer_sizes[ii]
-            a = self.activations[ii]
             self._layers.append(Dropout(self.dropout))
             self._layers.append(
                 GraphConvolution(
-                    l,
-                    activation=a,
+                    self.layer_sizes[ii],
+                    activation=self.activations[ii],
                     use_bias=self.bias,
-                    kernel_regularizer=self.kernel_regularizer,
                     final_layer=ii == (n_layers - 1),
+                    kernel_initializer=kernel_initializer,
+                    kernel_regularizer=kernel_regularizer,
+                    kernel_constraint=kernel_constraint,
+                    bias_initializer=bias_initializer,
+                    bias_regularizer=bias_regularizer,
+                    bias_constraint=bias_constraint,
                 )
             )
 
@@ -354,9 +382,9 @@ class GCN:
         if self.use_sparse:
             A_indices, A_values = As
             Ainput = [
-                SqueezedSparseConversion(shape=(n_nodes, n_nodes))(
-                    [A_indices, A_values]
-                )
+                SqueezedSparseConversion(
+                    shape=(n_nodes, n_nodes), dtype=A_values.dtype
+                )([A_indices, A_values])
             ]
 
         # Otherwise, create dense matrix from input tensor
@@ -370,35 +398,40 @@ class GCN:
             )
 
         h_layer = x_in
+        if self.method == "none":
+            # For GCN, if no preprocessing has been done, we apply the preprocessing layer to perform that.
+            Ainput = [self.graph_norm_layer(Ainput[0])]
         for layer in self._layers:
             if isinstance(layer, GraphConvolution):
                 # For a GCN layer add the matrix and output indices
                 # Note that the output indices are only used if `final_layer=True`
                 h_layer = layer([h_layer, out_indices] + Ainput)
-
             else:
                 # For other (non-graph) layers only supply the input tensor
                 h_layer = layer(h_layer)
 
-            # print("Hlayer:", h_layer)
-
         return h_layer
 
-    def node_model(self):
+    def in_out_tensors(self, multiplicity=None):
         """
-        Builds a GCN model for node prediction
+        Builds a GCN model for node or link prediction
 
         Returns:
-            tuple: `(x_inp, x_out)`, where `x_inp` is a list of two Keras input tensors for the GCN model (containing node features and graph laplacian),
-            and `x_out` is a Keras tensor for the GCN model output.
+            tuple: `(x_inp, x_out)`, where `x_inp` is a list of Keras/TensorFlow
+            input tensors for the GCN model and `x_out` is a tensor of the GCN model output.
         """
-        # Placeholder for node features
-        N_nodes = self.generator.features.shape[0]
-        N_feat = self.generator.features.shape[1]
+        # Inputs for features
+        x_t = Input(batch_shape=(1, self.n_nodes, self.n_features))
 
-        # Inputs for features & target indices
-        x_t = Input(batch_shape=(1, N_nodes, N_feat))
-        out_indices_t = Input(batch_shape=(1, None), dtype="int32")
+        # If not specified use multiplicity from instanciation
+        if multiplicity is None:
+            multiplicity = self.multiplicity
+
+        # Indices to gather for model output
+        if multiplicity == 1:
+            out_indices_t = Input(batch_shape=(1, None), dtype="int32")
+        else:
+            out_indices_t = Input(batch_shape=(1, None, multiplicity), dtype="int32")
 
         # Create inputs for sparse or dense matrices
         if self.use_sparse:
@@ -409,7 +442,7 @@ class GCN:
 
         else:
             # Placeholders for the dense adjacency matrix
-            A_m = Input(batch_shape=(1, N_nodes, N_nodes))
+            A_m = Input(batch_shape=(1, self.n_nodes, self.n_nodes))
             A_placeholders = [A_m]
 
         # TODO: Support multiple matrices
@@ -424,3 +457,21 @@ class GCN:
             self.x_out_flat = x_out
 
         return x_inp, x_out
+
+    def _link_model(self):
+        if self.multiplicity != 2:
+            warnings.warn(
+                "Link model requested but a generator not supporting links was supplied."
+            )
+        return self.in_out_tensors(multiplicity=2)
+
+    def _node_model(self):
+        if self.multiplicity != 1:
+            warnings.warn(
+                "Node model requested but a generator not supporting nodes was supplied."
+            )
+        return self.in_out_tensors(multiplicity=1)
+
+    node_model = deprecated_model_function(_node_model, "node_model")
+    link_model = deprecated_model_function(_link_model, "link_model")
+    build = deprecated_model_function(in_out_tensors, "build")
