@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright 2018-2019 Data61, CSIRO
+# Copyright 2018-2020 Data61, CSIRO
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,12 +20,13 @@ Definition of Graph Attention Network (GAT) layer, and GAT class that is a stack
 __all__ = ["GraphAttention", "GraphAttentionSparse", "GAT"]
 
 import warnings
-from keras import backend as K
-from keras import activations, constraints, initializers, regularizers
-from keras.layers import Input, Layer, Dropout, LeakyReLU, Lambda, Reshape
+import tensorflow as tf
+from tensorflow.keras import backend as K
+from tensorflow.keras import activations, constraints, initializers, regularizers
+from tensorflow.keras.layers import Input, Layer, Dropout, LeakyReLU, Lambda, Reshape
 
-from ..mapper import FullBatchNodeGenerator
-from .misc import SqueezedSparseConversion
+from ..mapper import FullBatchNodeGenerator, FullBatchGenerator
+from .misc import SqueezedSparseConversion, deprecated_model_function
 
 
 class GraphAttention(Layer):
@@ -65,17 +66,17 @@ class GraphAttention(Layer):
         final_layer (bool): If False the layer returns output for all nodes,
                             if True it returns the subset specified by the indices passed to it.
         use_bias (bool): toggles an optional bias
-        kernel_initializer (str): name of layer bias f the initializer for kernel parameters (weights)
-        bias_initializer (str): name of the initializer for bias
-        attn_kernel_initializer (str): name of the initializer for attention kernel
-        kernel_regularizer (str): name of regularizer to be applied to layer kernel. Must be a Keras regularizer.
-        bias_regularizer (str): name of regularizer to be applied to layer bias. Must be a Keras regularizer.
-        attn_kernel_regularizer (str): name of regularizer to be applied to attention kernel. Must be a Keras regularizer.
-        activity_regularizer (str): not used in the current implementation
-        kernel_constraint (str): constraint applied to layer's kernel. Must be a Keras constraint https://keras.io/constraints/
-        bias_constraint (str): constraint applied to layer's bias. Must be a Keras constraint https://keras.io/constraints/
-        attn_kernel_constraint (str): constraint applied to attention kernel. Must be a Keras constraint https://keras.io/constraints/
-        **kwargs: optional keyword arguments supplied to the Keras :class:`Layer`
+        saliency_map_support (bool): If calculating saliency maps using the tools in
+            stellargraph.interpretability.saliency_maps this should be True. Otherwise this should be False (default).
+        kernel_initializer (str or func, optional): The initialiser to use for the head weights.
+        kernel_regularizer (str or func, optional): The regulariser to use for the head weights.
+        kernel_constraint (str or func, optional): The constraint to use for the head weights.
+        bias_initializer (str or func, optional): The initialiser to use for the head bias.
+        bias_regularizer (str or func, optional): The regulariser to use for the head bias.
+        bias_constraint (str or func, optional): The constraint to use for the head bias.
+        attn_kernel_initializer (str or func, optional): The initialiser to use for the attention weights.
+        attn_kernel_regularizer (str or func, optional): The regulariser to use for the attention weights.
+        attn_kernel_constraint (str or func, optional): The constraint to use for the attention weights.
     """
 
     def __init__(
@@ -88,17 +89,17 @@ class GraphAttention(Layer):
         activation="relu",
         use_bias=True,
         final_layer=False,
+        saliency_map_support=False,
         kernel_initializer="glorot_uniform",
-        bias_initializer="zeros",
-        attn_kernel_initializer="glorot_uniform",
         kernel_regularizer=None,
-        bias_regularizer=None,
-        attn_kernel_regularizer=None,
-        activity_regularizer=None,
         kernel_constraint=None,
+        bias_initializer="zeros",
+        bias_regularizer=None,
         bias_constraint=None,
+        attn_kernel_initializer="glorot_uniform",
+        attn_kernel_regularizer=None,
         attn_kernel_constraint=None,
-        **kwargs
+        **kwargs,
     ):
 
         if attn_heads_reduction not in {"concat", "average"}:
@@ -117,19 +118,7 @@ class GraphAttention(Layer):
         self.use_bias = use_bias
         self.final_layer = final_layer
 
-        self.kernel_initializer = initializers.get(kernel_initializer)
-        self.bias_initializer = initializers.get(bias_initializer)
-        self.attn_kernel_initializer = initializers.get(attn_kernel_initializer)
-
-        self.kernel_regularizer = regularizers.get(kernel_regularizer)
-        self.bias_regularizer = regularizers.get(bias_regularizer)
-        self.attn_kernel_regularizer = regularizers.get(attn_kernel_regularizer)
-        self.activity_regularizer = regularizers.get(activity_regularizer)
-
-        self.kernel_constraint = constraints.get(kernel_constraint)
-        self.bias_constraint = constraints.get(bias_constraint)
-        self.attn_kernel_constraint = constraints.get(attn_kernel_constraint)
-
+        self.saliency_map_support = saliency_map_support
         # Populated by build()
         self.kernels = []  # Layer kernels for attention heads
         self.biases = []  # Layer biases for attention heads
@@ -142,6 +131,16 @@ class GraphAttention(Layer):
             # Output will have shape (..., F')
             self.output_dim = self.units
 
+        self.kernel_initializer = initializers.get(kernel_initializer)
+        self.kernel_regularizer = regularizers.get(kernel_regularizer)
+        self.kernel_constraint = constraints.get(kernel_constraint)
+        self.bias_initializer = initializers.get(bias_initializer)
+        self.bias_regularizer = regularizers.get(bias_regularizer)
+        self.bias_constraint = constraints.get(bias_constraint)
+        self.attn_kernel_initializer = initializers.get(attn_kernel_initializer)
+        self.attn_kernel_regularizer = regularizers.get(attn_kernel_regularizer)
+        self.attn_kernel_constraint = constraints.get(attn_kernel_constraint)
+
         super().__init__(**kwargs)
 
     def get_config(self):
@@ -158,18 +157,19 @@ class GraphAttention(Layer):
             "activation": activations.serialize(self.activation),
             "use_bias": self.use_bias,
             "final_layer": self.final_layer,
+            "saliency_map_support": self.saliency_map_support,
             "kernel_initializer": initializers.serialize(self.kernel_initializer),
+            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
+            "kernel_constraint": constraints.serialize(self.kernel_constraint),
             "bias_initializer": initializers.serialize(self.bias_initializer),
+            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
+            "bias_constraint": constraints.serialize(self.bias_constraint),
             "attn_kernel_initializer": initializers.serialize(
                 self.attn_kernel_initializer
             ),
-            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
-            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
             "attn_kernel_regularizer": regularizers.serialize(
                 self.attn_kernel_regularizer
             ),
-            "kernel_constraint": constraints.serialize(self.kernel_constraint),
-            "bias_constraint": constraints.serialize(self.bias_constraint),
             "attn_kernel_constraint": constraints.serialize(
                 self.attn_kernel_constraint
             ),
@@ -183,7 +183,7 @@ class GraphAttention(Layer):
         Assumes the following inputs:
 
         Args:
-            input_shape (tuple of ints)
+            input_shapes (tuple of ints)
                 Shape tuples can include None for free dimensions, instead of an integer.
 
         Returns:
@@ -197,18 +197,29 @@ class GraphAttention(Layer):
         else:
             out_dim = feature_shape[1]
 
-        return (batch_dim, out_dim, self.output_dim)
+        return batch_dim, out_dim, self.output_dim
 
     def build(self, input_shapes):
         """
         Builds the layer
 
         Args:
-            input_shape (list of int): shapes of the layer's inputs (node features and adjacency matrix)
+            input_shapes (list of int): shapes of the layer's inputs (node features and adjacency matrix)
 
         """
         feat_shape = input_shapes[0]
-        input_dim = feat_shape[-1]
+        input_dim = int(feat_shape[-1])
+
+        # Variables to support integrated gradients
+        self.delta = self.add_weight(
+            name="ig_delta", shape=(), trainable=False, initializer=initializers.ones()
+        )
+        self.non_exist_edge = self.add_weight(
+            name="ig_non_exist_edge",
+            shape=(),
+            trainable=False,
+            initializer=initializers.zeros(),
+        )
 
         # Initialize weights for each attention head
         for head in range(self.attn_heads):
@@ -279,6 +290,7 @@ class GraphAttention(Layer):
         X = inputs[0]  # Node features (1 x N x F)
         out_indices = inputs[1]  # output indices (1 x K)
         A = inputs[2]  # Adjacency matrix (N x N)
+        N = K.int_shape(A)[-1]
 
         batch_dim, n_nodes, _ = K.int_shape(X)
         if batch_dim != 1:
@@ -322,11 +334,22 @@ class GraphAttention(Layer):
             # YT: this only works for 'binary' A, not for 'weighted' A!
             # YT: if A does not have self-loops, the node itself will be masked, so A should have self-loops
             # YT: this is ensured by setting the diagonal elements of A tensor to 1 above
-            mask = -10e9 * (1.0 - A)
-            dense += mask
+            if not self.saliency_map_support:
+                mask = -10e9 * (1.0 - A)
+                dense += mask
+                dense = K.softmax(dense)  # (N x N), Eq. 3 of the paper
 
-            # Apply softmax to get attention coefficients
-            dense = K.softmax(dense, axis=1)  # (N x N), Eq. 3 of the paper
+            else:
+                # dense = dense - tf.reduce_max(dense)
+                # GAT with support for saliency calculations
+                W = (self.delta * A) * K.exp(
+                    dense - K.max(dense, axis=1, keepdims=True)
+                ) * (1 - self.non_exist_edge) + self.non_exist_edge * (
+                    A + self.delta * (tf.ones((N, N)) - A) + tf.eye(N)
+                ) * K.exp(
+                    dense - K.max(dense, axis=1, keepdims=True)
+                )
+                dense = W / K.sum(W, axis=1, keepdims=True)
 
             # Apply dropout to features and attention coefficients
             dropout_feat = Dropout(self.in_dropout_rate)(features)  # (N x F')
@@ -398,22 +421,19 @@ class GraphAttentionSparse(GraphAttention):
         final_layer (bool): If False the layer returns output for all nodes,
                             if True it returns the subset specified by the indices passed to it.
         use_bias (bool): toggles an optional bias
-        kernel_initializer (str): name of layer bias f the initializer for kernel parameters (weights)
-        bias_initializer (str): name of the initializer for bias
-        attn_kernel_initializer (str): name of the initializer for attention kernel
-        kernel_regularizer (str): name of regularizer to be applied to layer kernel. Must be a Keras regularizer.
-        bias_regularizer (str): name of regularizer to be applied to layer bias. Must be a Keras regularizer.
-        attn_kernel_regularizer (str): name of regularizer to be applied to attention kernel. Must be a Keras regularizer.
-        activity_regularizer (str): not used in the current implementation
-        kernel_constraint (str): constraint applied to layer's kernel. Must be a Keras constraint https://keras.io/constraints/
-        bias_constraint (str): constraint applied to layer's bias. Must be a Keras constraint https://keras.io/constraints/
-        attn_kernel_constraint (str): constraint applied to attention kernel. Must be a Keras constraint https://keras.io/constraints/
-        **kwargs: optional keyword arguments supplied to the Keras :class:`Layer`
+        saliency_map_support (bool): If calculating saliency maps using the tools in
+            stellargraph.interpretability.saliency_maps this should be True. Otherwise this should be False (default).
+        kernel_initializer (str or func, optional): The initialiser to use for the head weights.
+        kernel_regularizer (str or func, optional): The regulariser to use for the head weights.
+        kernel_constraint (str or func, optional): The constraint to use for the head weights.
+        bias_initializer (str or func, optional): The initialiser to use for the head bias.
+        bias_regularizer (str or func, optional): The regulariser to use for the head bias.
+        bias_constraint (str or func, optional): The constraint to use for the head bias.
+        attn_kernel_initializer (str or func, optional): The initialiser to use for the attention weights.
+        attn_kernel_regularizer (str or func, optional): The regulariser to use for the attention weights.
+        attn_kernel_constraint (str or func, optional): The constraint to use for the attention weights.
     """
 
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-
     def call(self, inputs, **kwargs):
         """
         Creates the layer as a Keras graph
@@ -435,7 +455,7 @@ class GraphAttentionSparse(GraphAttention):
         out_indices = inputs[1]  # output indices (1 x K)
         A_sparse = inputs[2]  # Adjacency matrix (1 x N x N)
 
-        if not isinstance(A_sparse, K.tf.SparseTensor):
+        if not isinstance(A_sparse, tf.SparseTensor):
             raise TypeError("A is not sparse")
 
         # Get undirected graph edges (E x 2)
@@ -470,16 +490,11 @@ class GraphAttentionSparse(GraphAttention):
                 features, attention_kernel[1]
             )  # (N x 1), [a_2]^T [Wh_j]
 
-            # Attention head a(Wh_i, Wh_j) = a^T [[Wh_i], [Wh_j]]
-            dense = attn_for_self + K.transpose(
-                attn_for_neighs
-            )  # (N x N) via broadcasting
-
             # Create sparse attention vector (All non-zero values of the matrix)
-            sparse_attn_self = K.tf.gather(
+            sparse_attn_self = tf.gather(
                 K.reshape(attn_for_self, [-1]), A_indices[:, 0], axis=0
             )
-            sparse_attn_neighs = K.tf.gather(
+            sparse_attn_neighs = tf.gather(
                 K.reshape(attn_for_neighs, [-1]), A_indices[:, 1], axis=0
             )
             attn_values = sparse_attn_self + sparse_attn_neighs
@@ -492,17 +507,17 @@ class GraphAttentionSparse(GraphAttention):
             dropout_attn = Dropout(self.attn_dropout_rate)(attn_values)  # (N x N)
 
             # Convert to sparse matrix
-            sparse_attn = K.tf.sparse.SparseTensor(
+            sparse_attn = tf.sparse.SparseTensor(
                 A_indices, values=dropout_attn, dense_shape=[n_nodes, n_nodes]
             )
 
             # Apply softmax to get attention coefficients
-            sparse_attn = K.tf.sparse.softmax(
-                sparse_attn
-            )  # (N x N), Eq. 3 of the paper
+            sparse_attn = tf.sparse.softmax(sparse_attn)  # (N x N), Eq. 3 of the paper
 
             # Linear combination with neighbors' features [YT: see Eq. 4]
-            node_features = K.tf.sparse.matmul(sparse_attn, dropout_feat)  # (N x F')
+            node_features = tf.sparse.sparse_dense_matmul(
+                sparse_attn, dropout_feat
+            )  # (N x F')
 
             if self.use_bias:
                 node_features = K.bias_add(node_features, self.biases[head])
@@ -528,19 +543,27 @@ class GraphAttentionSparse(GraphAttention):
         return output
 
 
+def _require_without_generator(value, name):
+    if value is not None:
+        return value
+    else:
+        raise ValueError(
+            f"{name}: expected a value for 'input_dim', 'node_num' and 'multiplicity' when "
+            f"'generator' is not provided, found {name}=None."
+        )
+
+
 class GAT:
     """
     A stack of Graph Attention (GAT) layers with aggregation of multiple attention heads,
     Eqs 5-6 of the GAT paper https://arxiv.org/abs/1710.10903
 
     To use this class as a Keras model, the features and pre-processed adjacency matrix
-    should be supplied using the :class:`FullBatchNodeGenerator` class. To have the appropriate
-    pre-processing the generator object should be instantiated as follows::
+    should be supplied using either the :class:`FullBatchNodeGenerator` class for node inference
+    or the :class:`FullBatchLinkGenerator` class for link inference.
 
-        generator = FullBatchNodeGenerator(G, sparse=False, method="gat")
-
-    For more details, please see the GAT demo notebook:
-    demos/node-classification/gat/gat-cora-node-classification-example.ipynb
+    To have the appropriate pre-processing the generator object should be instanciated
+    with the `method='gat'` argument.
 
     Examples:
         Creating a GAT node classification model from an existing :class:`StellarGraph` object `G`::
@@ -554,7 +577,10 @@ class GAT:
                     in_dropout=0.5,
                     attn_dropout=0.5,
                 )
-            x_inp, predictions = gat.node_model()
+            x_inp, predictions = gat.in_out_tensors()
+
+    For more details, please see the GAT demo notebook:
+    demos/node-classification/gat/gat-cora-node-classification-example.ipynb
 
     Notes:
       - The inputs are tensors with a batch dimension of 1. These are provided by the \
@@ -572,6 +598,7 @@ class GAT:
     Args:
         layer_sizes (list of int): list of output sizes of GAT layers in the stack. The length of this list defines
             the number of GraphAttention layers in the stack.
+        generator (FullBatchNodeGenerator): an instance of FullBatchNodeGenerator class constructed on the graph of interest
         attn_heads (int or list of int): number of attention heads in GraphAttention layers. The options are:
 
             - a single integer: the passed value of ``attn_heads`` will be applied to all GraphAttention layers in the stack, except the last layer (for which the number of attn_heads will be set to 1).
@@ -581,30 +608,63 @@ class GAT:
             for all layers in the stack. Valid entries in the list are {'concat', 'average'}.
             If None is passed, the default reductions are applied: 'concat' reduction to all layers in the stack
             except the final layer, 'average' reduction to the last layer (Eqs. 5-6 of the GAT paper).
-        activations (list of str): list of activations applied to each layer's output
         bias (bool): toggles an optional bias in GAT layers
         in_dropout (float): dropout rate applied to input features of each GAT layer
         attn_dropout (float): dropout rate applied to attention maps
         normalize (str or None): normalization applied to the final output features of the GAT layers stack. Default is None.
-        generator (FullBatchNodeGenerator): an instance of FullBatchNodeGenerator class constructed on the graph of interest
+        activations (list of str): list of activations applied to each layer's output; defaults to ['elu', ..., 'elu'].
+        saliency_map_support (bool): If calculating saliency maps using the tools in
+            stellargraph.interpretability.saliency_maps this should be True. Otherwise this should be False (default).
+        multiplicity (int, optional): The number of nodes to process at a time. This is 1 for a node
+            inference and 2 for link inference (currently no others are supported).
+        num_nodes (int, optional): The number of nodes in the given graph.
+        num_features (int, optional): The dimensions of the node features used as input to the model.
+        kernel_initializer (str or func, optional): The initialiser to use for the weights of each layer.
+        kernel_regularizer (str or func, optional): The regulariser to use for the weights of each layer.
+        kernel_constraint (str or func, optional): The constraint to use for the weights of each layer.
+        bias_initializer (str or func, optional): The initialiser to use for the bias of each layer.
+        bias_regularizer (str or func, optional): The regulariser to use for the bias of each layer.
+        bias_constraint (str or func, optional): The constraint to use for the bias of each layer.
+        attn_kernel_initializer (str or func, optional): The initialiser to use for the attention weights.
+        attn_kernel_regularizer (str or func, optional): The regulariser to use for the attention weights.
+        attn_kernel_constraint (str or func, optional): The constraint to use for the attention bias.
+
+    .. note::
+        The values for ``multiplicity``, ``num_nodes``, and ``num_features`` are obtained from the
+        provided ``generator`` by default. The additional keyword arguments for these parameters
+        provide an alternative way to specify them if a generator cannot be supplied.
     """
 
     def __init__(
         self,
         layer_sizes,
-        activations,
+        generator=None,
         attn_heads=1,
         attn_heads_reduction=None,
         bias=True,
         in_dropout=0.0,
         attn_dropout=0.0,
         normalize=None,
-        generator=None,
+        activations=None,
+        saliency_map_support=False,
+        multiplicity=1,
+        num_nodes=None,
+        num_features=None,
+        kernel_initializer="glorot_uniform",
+        kernel_regularizer=None,
+        kernel_constraint=None,
+        bias_initializer="zeros",
+        bias_regularizer=None,
+        bias_constraint=None,
+        attn_kernel_initializer="glorot_uniform",
+        attn_kernel_regularizer=None,
+        attn_kernel_constraint=None,
     ):
         self.bias = bias
         self.in_dropout = in_dropout
         self.attn_dropout = attn_dropout
         self.generator = generator
+        self.saliency_map_support = saliency_map_support
 
         # Check layer_sizes (must be list of int):
         # check type:
@@ -622,14 +682,15 @@ class GAT:
                 )
             )
         self.layer_sizes = layer_sizes
+        n_layers = len(layer_sizes)
 
         # Check attn_heads (must be int or list of int):
         if isinstance(attn_heads, list):
             # check the length
-            if not len(attn_heads) == len(layer_sizes):
+            if not len(attn_heads) == n_layers:
                 raise ValueError(
                     "{}: length of attn_heads list ({}) should match the number of GAT layers ({})".format(
-                        type(self).__name__, len(attn_heads), len(layer_sizes)
+                        type(self).__name__, len(attn_heads), n_layers
                     )
                 )
             # check that values in the list are valid
@@ -645,7 +706,7 @@ class GAT:
             self.attn_heads = list()
             for l, _ in enumerate(layer_sizes):
                 # number of attention heads for layer l: attn_heads (int) for all but the last layer (for which it's set to 1)
-                self.attn_heads.append(attn_heads if l < len(layer_sizes) - 1 else 1)
+                self.attn_heads.append(attn_heads if l < n_layers - 1 else 1)
 
         else:
             raise TypeError(
@@ -657,9 +718,7 @@ class GAT:
         # Check attn_heads_reduction (list of str, or None):
         if attn_heads_reduction is None:
             # set default head reductions, see eqs 5-6 of the GAT paper
-            self.attn_heads_reduction = ["concat"] * (len(layer_sizes) - 1) + [
-                "average"
-            ]
+            self.attn_heads_reduction = ["concat"] * (n_layers - 1) + ["average"]
         else:
             # user-specified list of head reductions (valid entries are 'concat' and 'average')
             # check type (must be a list of str):
@@ -674,7 +733,7 @@ class GAT:
             if not len(attn_heads_reduction) == len(layer_sizes):
                 raise ValueError(
                     "{}: length of attn_heads_reduction list ({}) should match the number of GAT layers ({})".format(
-                        type(self).__name__, len(attn_heads_reduction), len(layer_sizes)
+                        type(self).__name__, len(attn_heads_reduction), n_layers
                     )
                 )
 
@@ -692,6 +751,8 @@ class GAT:
 
         # Check activations (list of str):
         # check type:
+        if activations is None:
+            activations = ["elu"] * n_layers
         if not isinstance(activations, list):
             raise TypeError(
                 "{}: activations should be a list of strings; received {} instead".format(
@@ -699,28 +760,37 @@ class GAT:
                 )
             )
         # check length:
-        if not len(activations) == len(layer_sizes):
+        if not len(activations) == n_layers:
             raise ValueError(
                 "{}: length of activations list ({}) should match the number of GAT layers ({})".format(
-                    type(self).__name__, len(activations), len(layer_sizes)
+                    type(self).__name__, len(activations), n_layers
                 )
             )
         self.activations = activations
 
-        # check generator:
-        if generator is not None:
-            if not isinstance(generator, FullBatchNodeGenerator):
-                raise ValueError(
-                    "{}: generator must be of type FullBatchNodeGenerator or None; received object of type {} instead".format(
-                        type(self).__name__, type(generator).__name__
-                    )
-                )
-
-            # Check if the generator is producing a sparse matrix
-            self.use_sparse = generator.use_sparse
+        # Check generator and configure sparse adjacency matrix
+        if generator is None:
+            self.use_sparse = False
+            self.multiplicity = _require_without_generator(multiplicity, "multiplicity")
+            self.n_nodes = _require_without_generator(num_nodes, "num_nodes")
+            self.n_features = _require_without_generator(num_features, "num_features")
 
         else:
-            self.use_sparse = False
+            if not isinstance(generator, FullBatchGenerator):
+                raise TypeError(
+                    "Generator should be a instance of FullBatchNodeGenerator or FullBatchLinkGenerator"
+                )
+
+            # Copy required information from generator
+            self.use_sparse = generator.use_sparse
+            self.multiplicity = generator.multiplicity
+            self.n_nodes = generator.features.shape[0]
+            self.n_features = generator.features.shape[1]
+
+        if self.n_nodes is None or self.n_features is None:
+            raise RuntimeError(
+                "node_model: if generator is not provided to object constructor, num_nodes and feature_size must be specified."
+            )
 
         # Set the normalization layer used in the model
         if normalize == "l2":
@@ -760,6 +830,16 @@ class GAT:
                     activation=self.activations[ii],
                     use_bias=self.bias,
                     final_layer=ii == (n_layers - 1),
+                    saliency_map_support=self.saliency_map_support,
+                    kernel_initializer=kernel_initializer,
+                    kernel_regularizer=kernel_regularizer,
+                    kernel_constraint=kernel_constraint,
+                    bias_initializer=bias_initializer,
+                    bias_regularizer=bias_regularizer,
+                    bias_constraint=bias_constraint,
+                    attn_kernel_initializer=attn_kernel_initializer,
+                    attn_kernel_regularizer=attn_kernel_regularizer,
+                    attn_kernel_constraint=attn_kernel_constraint,
                 )
             )
 
@@ -773,13 +853,14 @@ class GAT:
         Returns: Output tensor of the GAT layers stack
 
         """
-        assert isinstance(inputs, list), "input must be a list, got {} instead".format(
-            type(inputs)
-        )
+        if not isinstance(inputs, list):
+            raise TypeError(f"inputs: expected list, found {type(inputs).__name__}")
+
         x_in, out_indices, *As = inputs
 
         # Currently we require the batch dimension to be one for full-batch methods
         batch_dim, n_nodes, _ = K.int_shape(x_in)
+
         if batch_dim != 1:
             raise ValueError(
                 "Currently full-batch methods only support a batch dimension of one"
@@ -820,32 +901,27 @@ class GAT:
 
         return self._normalization(h_layer)
 
-    def node_model(self, num_nodes=None, feature_size=None):
+    def in_out_tensors(self, multiplicity=None):
         """
-        Builds a GCN model for node prediction
+        Builds a GAT model for node or link prediction
 
         Returns:
-            tuple: `(x_inp, x_out)`, where `x_inp` is a list of two Keras input tensors for the GCN model (containing node features and graph laplacian),
-            and `x_out` is a Keras tensor for the GCN model output.
+            tuple: `(x_inp, x_out)`, where `x_inp` is a list of Keras/TensorFlow
+            input tensors for the model and `x_out` is a tensor of the model output.
         """
-        # Create input tensor:
-        if self.generator is not None:
-            # Placeholder for node features
-            N_nodes = self.generator.features.shape[0]
-            N_feat = self.generator.features.shape[1]
 
-        elif num_nodes is not None and feature_size is not None:
-            N_nodes = num_nodes
-            N_feat = feature_size
+        # Inputs for features
+        x_t = Input(batch_shape=(1, self.n_nodes, self.n_features))
 
+        # If not specified use multiplicity from instanciation
+        if multiplicity is None:
+            multiplicity = self.multiplicity
+
+        # Indices to gather for model output
+        if multiplicity == 1:
+            out_indices_t = Input(batch_shape=(1, None), dtype="int32")
         else:
-            raise RuntimeError(
-                "node_model: if generator is not provided to object constructor, num_nodes and feature_size must be specified."
-            )
-
-        # Inputs for features & target indices
-        x_t = Input(batch_shape=(1, N_nodes, N_feat))
-        out_indices_t = Input(batch_shape=(1, None), dtype="int32")
+            out_indices_t = Input(batch_shape=(1, None, multiplicity), dtype="int32")
 
         # Create inputs for sparse or dense matrices
         if self.use_sparse:
@@ -856,10 +932,10 @@ class GAT:
 
         else:
             # Placeholders for the dense adjacency matrix
-            A_m = Input(batch_shape=(1, N_nodes, N_nodes))
+            A_m = Input(batch_shape=(1, self.n_nodes, self.n_nodes))
             A_placeholders = [A_m]
 
-        # TODO: Support multiple matrices?
+        # TODO: Support multiple matrices
         x_inp = [x_t, out_indices_t] + A_placeholders
         x_out = self(x_inp)
 
@@ -871,17 +947,20 @@ class GAT:
 
         return x_inp, x_out
 
-    def link_model(self):
-        """
-        Builds a GAT model for link (node pair) prediction (implementation pending)
+    def _link_model(self):
+        if self.multiplicity != 2:
+            warnings.warn(
+                "Link model requested but a generator not supporting links was supplied."
+            )
+        return self.in_out_tensors(multiplicity=2)
 
-        """
-        raise NotImplemented
+    def _node_model(self):
+        if self.multiplicity != 1:
+            warnings.warn(
+                "Node model requested but a generator not supporting nodes was supplied."
+            )
+        return self.in_out_tensors(multiplicity=1)
 
-    def default_model(self, flatten_output=False):
-        warnings.warn(
-            "The .default_model() method will be deprecated in future versions. "
-            "Please use .node_model() or .link_model() methods instead.",
-            PendingDeprecationWarning,
-        )
-        return self.node_model()
+    node_model = deprecated_model_function(_node_model, "node_model")
+    link_model = deprecated_model_function(_link_model, "link_model")
+    build = deprecated_model_function(in_out_tensors, "build")
